var Assert = require('@kingjs/assert');

var defineConst = { configurable: false, enumerable: true, writable: false }
var defineConstHidden = { configurable: false, enumerable: false, writable: false }

// defineProperty(target, name, initialize, configuration)
Object.defineProperty(
  this, 'defineProperty', Object.create(defineConstHidden, { 
    value: {
      enumerable: true,
      value: function(target, name, initialize, configuration) {
        target[name] = initialize;
        Object.defineProperty(target, name, configuration);
        return initialize;
      }
    }
  })
);
var defineProperty = this.defineProperty;

// defineConstHiddenProperty(target, name, value)
defineProperty(this, 'defineConstHiddenProperty', function(target, name, value) {
  return defineProperty(target, name, value, defineConstHidden);
}, defineConstHidden);
var defineConstHiddenProperty = this.defineConstHiddenProperty;

// defineFunction(target, func)
defineConstHiddenProperty(this, 'defineFunction', function(target, func) {
  return defineConstHiddenProperty(target, func.name, func);
});
var defineFunction = this.defineFunction;

// defineFunctions(target, funcs)
defineFunction(this, function defineFunctions(target, funcs) {
  for (name in funcs)
    defineConstHiddenProperty(target, name, funcs[name]);
})
var defineFunctions = this.defineFunctions;

// defineConstProperty(target)
// defineConstProperties(target, properties)
defineFunctions(this, { 
    
  defineConstProperty: function(target, name, value) {
    return defineProperty(target, name, value, defineConst);
  },
  
  defineConstProperties: function(target, properties) {
    for (name in properties)
      defineConstProperty(target, name, properties[name]);
  },
});
var defineConstProperty = this.defineConstProperty;
var defineConstProperties = this.defineConstProperties;

// createHashtable
// freeze(target)
defineFunctions(this, { 
    
  createHashtable: function() {
    // so: 'toString' in createHashtable() => false
    return Object.create(null);
  },
  
  freeze: function(target) {
    if (typeof target == 'object' || typeof target == 'function')
      Object.freeze(target);
    return target;
  },
});
var createHashtable = this.createHashtable;
var freeze = this.freeze;

defineConstProperties(exports, {
  emptyString: freeze(''),
  emptyArray: freeze([ ]),
  emptyObject: freeze({ }),
  emptyHashtable: freeze(createHashtable())
});

// Array.find(predicate)
defineFunction(Array.prototype, function find(predicate) {
  for (var i = 0; i < this.length; i ++) {
    var value = this[i];
    if (predicate(value))
        return value;
  }
});

// String.endsWith(value)
defineFunction(String.prototype, function endsWith(value) {
  var i = this.length;
  var j = value.length;

  if (j > i)
    return false;
  
  while (j >= 0 & i >= 0) {
    if (this[i--] != value[j--])
      return false;
  }
  
  return true;
});

// findValue()
// bindProxy()
// toSet()
// allKeys(source)
// replaceEach(func, array)
// pluck(target, name)
// scorch(target)
// rename(target, name, newName)
// create(proto)
defineFunctions(this, { 

  findValue: function(source, predicate) {
    for (var name in source) {
      var value = source[name];
      if (predicate(value))
        return value;
    }
  },

  bindProxy: function(func, proxy) {
    return function() { 
      Array.prototype.unshift.call(arguments, func);
      return proxy.apply(this, arguments); 
    };
  },

  toSet: function() {
    var result = createHashtable();
    
    for (var i = 0; i < arguments.length; i++)
      result[arguments[i]] = undefined;
    
    return result;
  },
  
  allKeys: function(source) {
    var result = [];
    for (var key in source)
      result.push(key);
    return result;
  },
  
  replaceEach: function(func, array) {
    for (var i = 0; i < array.length; i++)
      array[i] = func(array[i]);
  },
  
  pluck: function(target, name) {
    var value = target[name];
    delete target[name];
    return value;
  },
 
  scorch: function (target) {
    for (var name in target) {
      if (target[name] === undefined)
        delete target[name];
    }
    
    return target;
  },
  
  rename: function(target, name, newName) {
    if (name == newName)
      return newName;
    
    Assert.that(delete target[newName]);
    
    target[newName] = target[name];
    delete target[name];
    
    return newName;
  },
  
  create: function(proto) {
    
    if (proto == Array)
      return [ ];
    
    if (proto == String)
      return "";
    
    return Object.create(proto || Object.prototype);
  },
});