var Assert = require('@kingjs/assert');

var Bootstrap = require('./bootstrap');
var defineConst = Bootstrap.defineConst;
var defineConstHidden = Bootstrap.defineConstHidden;
var defineConstProperty = Bootstrap.defineConstProperty;
var bindProxy = Bootstrap.bindProxy;
var findValue = Bootstrap.findValue;

var Map = require('./map');
var mapDescriptors = Map.mapDescriptors;

var Assign = require('./assign');
var assign = Assign.assign;

var Operators = require('./operators');
var bind1st = Operators.bind1st;

var Reflection = require('./reflection');
var isFunction = Reflection.isFunction;
var isNamedFunction = Reflection.isNamedFunction;
var isObjectLiteral = Reflection.isObjectLiteral;
var isString = Reflection.isString;

var Str = require('./string');
var capitalize = Str.capitalize;

var Plural = require('./plural');
var defineSingularPlural = Plural.defineSingularPlural;
var pluralOf = Plural.pluralOf;

// default plurals
mapDescriptors(defineSingularPlural, {
  property: 'properties',
  // ...
})

function defineConfigurations(type, configurations, normalizeArgumentsProxy) {

  // define variations on Object.defineProperty and Object.defineProperties
  // for the property type with different names for each configuration combination
  mapDescriptors(
    function(name, configuration) {
      name += type;
      var defineConfiguredProperty = bindProxy(
        function(target, name, descriptor) {
          
          // merge configuration into descriptor 
          descriptor = assign(descriptor, configuration);
          
          if (target)
            Object.defineProperty.call(this, target, name, descriptor);
          
          return descriptor;     
        },
        normalizeArgumentsProxy
      );
      
      defineConfigurationVariants(name, defineConfiguredProperty);
    }, configurations
  );  
}

function defineConfigurationVariants(name, func) {
  
  // (define|set)[Const][Hidden](Property|Accessor)(target, name, descriptor);
  defineConstProperty(exports, name, func);
  
  // (define|set)[Const][Hidden](Properties|Accessors)(target, descriptors)
  var pluralName = pluralOf(name);
  var pluralFunc = defineConstProperty(exports, pluralName, function(fixedTarget, descriptors) {
    return mapDescriptors(bind1st(func, fixedTarget), descriptors);
  });
  
  var onTargets = 'OnTargets';
  
  // (define|set)[Const][Hidden](Property|Accessor)OnTargets(targets, name, descriptor)
  var nameOnTargets = name + onTargets;
  defineConstProperty(exports, nameOnTargets, function(targets, fixedName, fixedDescriptor, fixedSetter) {
    return targets.map(function(x) { return func(x, fixedName, fixedDescriptor, fixedSetter); }, this);
  });
  
  // (define|set)[Const][Hidden](Properties|Accessors)OnTargets(targets, descriptor)
  var pluralNameOnTargets = pluralName + onTargets;
  defineConstProperty(exports, pluralNameOnTargets, function(targets, fixedDescriptor) {
    return targets.map(function(x) { return pluralFunc(x, fixedDescriptor); }, this)
  });
};

// defineFunction = function(target, name, descriptor) { ... }
// defineFunctionOnTargets = function(targets, name, descriptor) { ... }
// defineFunctions = function(target, descriptors) { ... }
// defineFunctionsOnTargets = function(targets, descriptors) { ... }

// setProperty = function(target, name, descriptor) { ... }
// setConstProperty = function(target, name, descriptor) { ... }
// setHiddenProperty = function(target, name, descriptor) { ... }
// setConstHiddenProperty = function(target, name, descriptor) { ... }
// defineProperty = function(target, name, descriptor) { ... }
// defineConstProperty = function(target, name, descriptor) { ... }
// defineHiddenProperty = function(target, name, descriptor) { ... }
// defineConstHiddenProperty = function(target, name, descriptor) { ... }
// setPropertyOnTargets = function(targets, name, descriptor) { ... }
// setConstPropertyOnTargets = function(targets, name, descriptor) { ... }
// setHiddenPropertyOnTargets = function(targets, name, descriptor) { ... }
// setConstHiddenPropertyOnTargets = function(targets, name, descriptor) { ... }
// definePropertyOnTargets = function(targets, name, descriptor) { ... }
// defineConstPropertyOnTargets = function(targets, name, descriptor) { ... }
// defineHiddenPropertyOnTargets = function(targets, name, descriptor) { ... }
// defineConstHiddenPropertyOnTargets = function(targets, name, descriptor) { ... }
// setProperties = function(target, descriptors) { ... }
// setConstProperties = function(target, descriptors) { ... }
// setHiddenProperties = function(target, descriptors) { ... }
// setConstHiddenProperties = function(target, descriptors) { ... }
// defineProperties = function(target, descriptors) { ... }
// defineConstProperties = function(target, descriptors) { ... }
// defineHiddenProperties = function(target, descriptors) { ... }
// defineConstHiddenProperties = function(target, descriptors) { ... }
// setPropertiesOnTargets = function(targets, descriptors) { ... }
// setConstPropertiesOnTargets = function(targets, descriptors) { ... }
// setHiddenPropertiesOnTargets = function(targets, descriptors) { ... }
// setConstHiddenPropertiesOnTargets = function(targets, descriptors) { ... }
// definePropertiesOnTargets = function(targets, descriptors) { ... }
// defineConstPropertiesOnTargets = function(targets, descriptors) { ... }
// defineHiddenPropertiesOnTargets = function(targets, descriptors) { ... }
// defineConstHiddenPropertiesOnTargets = function(targets, descriptors) { ... }

// setAccessor = function(target, property, descriptor) { ... }
// setHiddenAccessor = function(target, property, descriptor) { ... }
// defineAccessor = function(target, property, descriptor) { ... }
// defineHiddenAccessor = function(target, property, descriptor) { ... }
// setAccessorOnTargets = function(targets, property, descriptor) { ... }
// setHiddenAccessorOnTargets = function(targets, property, descriptor) { ... }
// defineAccessorOnTargets = function(targets, property, descriptor) { ... }
// defineHiddenAccessorOnTargets = function(targets, property, descriptor) { ... }
// setAccessors = function(target, descriptors) { ... }
// setHiddenAccessors = function(target, descriptors) { ... }
// defineAccessors = function(target, descriptors) { ... }
// defineHiddenAccessors = function(target, descriptors) { ... }
// setAccessorsOnTargets = function(targets, descriptors) { ... }
// setHiddenAccessorsOnTargets = function(targets, descriptors) { ... }
// defineAccessorsOnTargets = function(targets, descriptors) { ... }
// defineHiddenAccessorsOnTargets = function(targets, descriptors) { ... }

mapDescriptors(
  function(name, descriptor) {    
    defineConfigurations(
      capitalize(name),
        descriptor.configurations,
        descriptor.normalizeArgumentsProxy
      );     
  }, 
  {
    'function': {
      configurations: {
        define: { configurable: false, enumerable: false, writable: false }
      }, 
      normalizeArgumentsProxy: function(next, target, name, descriptor) {
        
        // e.g. function foo() { ... } => 'foo', function foo() { ... }
        if (isNamedFunction(name)) {
          descriptor = name;
          name = descriptor.name;
        }
        
        // e.g. 'foo', function() { ... } => 'foo', { value: function() { ... } }
        if (isFunction(descriptor))
          descriptor = { value: descriptor };
        
        return next.call(this, target, name, descriptor);
      }
    },
    
    property: {
      configurations: { 
        set: { configurable: true, enumerable: true, writable: true },
        setConst: { configurable: true, enumerable: true, writable: false },
        setHidden: { configurable: true, enumerable: false, writable: true },
        setConstHidden: { configurable: true, enumerable: false, writable: false },
        define: { configurable: false, enumerable: true, writable: true },
        defineConst: defineConst,
        defineHidden: { configurable: false, enumerable: false, writable: true },
        defineConstHidden: defineConstHidden
      }, 
      normalizeArgumentsProxy: function(next, target, name, value) {
        
        // e.g. foo => { value: foo }
        return next.call(this, target, name, { value: value });
      }
    },
    
    accessor: {
      configurations: { 
        set: { configurable: true, enumerable: true },
        setHidden: { configurable: true, enumerable: false },
        define: { configurable: false, enumerable: true },
        defineHidden: { configurable: false, enumerable: false },
      },
      normalizeArgumentsProxy: function(next, target, name, descriptor) {
        
        // e.g. { get: function foo() { ... } } => 'foo', { get: function foo() { ... } }
        if (isObjectLiteral(name)) {
          descriptor = name;
          name = findValue(descriptor, isNamedFunction).name;
        }
        
        else if (!isObjectLiteral(descriptor)) {
          var get, set;
          
          // e.g. function foo() { ... } => 'foo', { get: function foo() { ... } }
          if (isFunction(name) || name == undefined) {
            get = name;
            set = descriptor;
            name = (get || set).name;
          } 
          
          // e.g. 'foo', function() { ... } => 'foo', { get: function() { ... } }
          else {
            get = descriptor;
            set = arguments[4];
          }
          
          descriptor = { };
          if (get) descriptor.get = get;
          if (set) descriptor.set = set;
        }
        
        Assert.that(isString(name));
        Assert.that(descriptor.get || descriptor.set);
        Assert.that(!descriptor.get || isFunction(descriptor.get));
        Assert.that(!descriptor.set || isFunction(descriptor.set));
        
        return next.call(this, target, name, descriptor);        
      }
    }
  }
);