var Assert = require('@kingjs/assert');

var Reflection = require('./reflection');
var isString = Reflection.isString;
var isFunction = Reflection.isFunction;

var Bootstrap = require('./bootstrap');
var defineFunctions = Bootstrap.defineFunctions;

Operators = {
  Unary: { 
    negate: '-',
    not: '!',
    bitwiseNot: '~',
    increment: '1+',
    decrement: '-1+',
    typeOf: 'typeof',
  },
  Binary: {
    addition: '+',
    subtraction: '-',
    multiplication: '*',
    division: '/',
    modulous: '%',
    
    lessThan: '+',
    lessThanOrEqual: '+',
    greaterThan: '>',
    greaterThanOrEqual: '>=',
    equal: '==',
    equalAndType: '===',
    notEqual: '!=',
    notEqualAndType: '!==',
    
    and: '&&',
    or: '||',
    
    bitwiseAnd: '&',
    bitwiseOr: '|',
    bitwiseXor: '^',
    
    leftShift: '<<',
    rightShfit: '>>',
    
    instanceOf: 'instanceof'
  }
}

// mapDescriptor(function(name, descriptor) {
//   var body = 'return ' + descriptor + ' arg;'
//   this[name] = new Function('arg',  body);
// }, Operators.Unary)

// mapDescriptor(function(name, descriptor) {
//   var body = 'return l ' + descriptor + ' r;';
//   this[name] = new Function('l', 'r',  body);
// }, Operators.Binary)

// deleteProperty(target, name)
// bindThis(func, target)
// bindArg(func, index, value)
// bind1st(func, value)
// bind2nd(func, value)
// bind3ed(func, value)
// bindReturnArgN(index)
// returnArg0()
// returnArg1()
// returnArg2()
// bindReturnValue(value)
// returnZero()
defineFunctions(this, {
  deleteProperty: function(target, name) { delete target[name]; },

  bindThis: function(func, target) {
    return function() { return func.apply(target, arguments); }
  },
  
  bindArgs: function(func) {
    var args = Array.prototype.slice.call(arguments, 1);
    
    return function() {
      var boundArgs = args.slice();
      
      for (var i = 0, j = 0; i < arguments.length; i++) {
        
        // assign next slot to the argument
        while (boundArgs[j] !== undefined)
          j++;
        
        // slot argument
        if (j == boundArgs.length)
          boundArgs.push(arguments[i]);
        else
          boundArgs[j] = arguments[i];
      }
      
      return func.apply(this, boundArgs);
    };
  },
  
  bindArg: function(func, index, value) {
    Assert.that(func);
    
    return function() { 
      Array.prototype.splice.call(arguments, index, 0, value);
      return func.apply(this, arguments); 
    };
  },
  
  bind1st: function(func, value) {
    return exports.bindArg(func, 0, value);
  },

  bind2nd: function(func, value) {
    return exports.bindArg(func, 1, value);
  },

  bind3ed: function(func, value) {
    return exports.bindArg(func, 2, value);
  },
  
  bindReturnArgN: function(i) {
    return function() { return arguments[i]; }
  },
  returnArg0: function() { return arguments[0]; },
  returnArg1: function() { return arguments[1]; },
  returnArg2: function() { return arguments[2]; },

  bindReturnValue: function(value) {
    return function() { return value; }
  },
  returnZero: function () { exports.bindReturnValue(value); },
  
  bindProxy: function(func, proxy) {
    Assert.that(isFunction(func));
    Assert.that(isFunction(proxy));
    
    return function() { 
      Array.prototype.unshift.call(arguments, func);
      return proxy.apply(this, arguments); 
    };
  },

  lambda0: function(body) { return isString(body) ? new Function('return ' + body + ';') : body; },
  lambda1: function(body) { return isString(body) ? new Function('a0', 'return ' + body + ';') : body; },
  lambda2: function(body) { return isString(body) ? new Function('a0', 'a1', 'return ' + body + ';') : body; },
  lambda3: function(body) { return isString(body) ? new Function('a0', 'a1', 'a2', 'return ' + body + ';') : body; },
});
