var Reflection = require('./reflection');
var isString = Reflection.isString;

var Bootstrap = require('./bootstrap');
var defineFunctions = Bootstrap.defineFunctions;
var allKeys = Bootstrap.allKeys;
var createHashtable = Bootstrap.createHashtable;
var emptyArray = Bootstrap.emptyArray;
var emptyHashtable = Bootstrap.emptyHashtable;

var Assign = require('./assign')
var assignDefaults = Assign.assignDefaults;
    
function walkPoset_(visited, action, vertices, edges, vertexName) {
  if (vertexName in visited)
    return;
  visited[vertexName] = undefined;

  var adjacentVertexNames = edges[vertexName] || emptyArray;
  for (var i = 0; i < adjacentVertexNames.length; i++) {
    var adjacentVertexName = adjacentVertexNames[i];
    if (isString(adjacentVertexName) && adjacentVertexName in vertices)
      walkPoset_(visited, action, vertices, edges, adjacentVertexName);
  }
  
  action(vertexName, adjacentVertexNames);
}

// walkPoset(action, vertices, edges)
// encodePoset(poset)
// decodePoset(poset)
defineFunctions(this, { 
  
  walkPoset: function(action, vertices, edges) {
    var visited = createHashtable();
    for (var vertexName in vertices)
      walkPoset_(visited, action, vertices, edges || emptyHashtable, vertexName); 
  },
  
  encodePoset: function(vertices, edges) {
    var poset = { };
    
    for (var vertexName in vertices) {
      var encodedVertexName = vertexName;
      var adjacentVertices = edges[vertexName];
      if (adjacentVertices && adjacentVertices.length > 0)
        encodedVertexName += '$' + adjacentVertices.join('$');
      poset[encodedVertexName] = vertices[vertexName];
    }
    
    return poset;
  },
  
  decodePoset: function(poset) {
    var vertices = createHashtable();
    var edges = createHashtable();
         
    // decode poset into vertices and edges
    // e.g. poset: { z: { c:2 }, y$z: { b:1 } } => 
    //  vertices: { z: { c:2 }, y: { b:1 } }
    //  edges: { z: [ ], y: [ 'z' ] }, 
    allKeys(poset).forEach(function(encodedVertexName) {    
      if (encodedVertexName.indexOf('$') == -1) {
        vertices[encodedVertexName] = poset[encodedVertexName];
        return;
      }
      
      var adjacentVertices = encodedVertexName.split('$');
      var vertexName = adjacentVertices.shift();
      
      vertices[vertexName] = poset[encodedVertexName];
      edges[vertexName] = adjacentVertices;
    });
    
    return {
      vertices: vertices,
      edges: edges
    };
  },
});
var walkPoset = this.walkPoset;
var decodePoset = this.decodePoset;

// walkEncodedPoset(action, poset)
// flattenEncodedPoset(poset)
defineFunctions(this, {

  walkEncodedPoset: function(action, poset) {
    poset = decodePoset(poset);
    walkPoset(function(vertexName, adjacentVertexNames) {
      action(poset.vertices, vertexName, adjacentVertexNames);
    }, poset.vertices, poset.edges);
  },
  
  // e.g. { z: { c:2 }, y$z: { b:1 } } => { z: { c:2 }, y: { c:2, b:1 } }
  flattenEncodedPoset: function(poset) {   
    poset = decodePoset(poset);
    var vertices = poset.vertices;
    var edges = poset.edges;
    
    // copy vertex properties to adjacent vertices if not preset already
    // e.g. previous example => vertices: { z: { c:2 }, y: { c:2, b:1 } }
    walkPoset(function(vertexName) {
      var vertex = vertices[vertexName];
      
      var adjacentVertexNames = edges[vertexName];
      if (adjacentVertexNames && adjacentVertexNames.length > 0) {
        adjacentVertexNames.forEach(function(adjacentVertexName) { 
          assignDefaults(vertex, vertices[adjacentVertexName]);
        });
      }
    }, vertices, edges);
    
    return vertices;
  },
});